# coding: utf-8

"""
    PKS

    PKS API  # noqa: E501

    OpenAPI spec version: 1.1.0
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


import pprint
import re  # noqa: F401

import six


class UpgradeClustersRequest(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'clusters': 'list[str]',
        'canaries': 'list[str]',
        'max_in_flight': 'int'
    }

    attribute_map = {
        'clusters': 'clusters',
        'canaries': 'canaries',
        'max_in_flight': 'max_in_flight'
    }

    def __init__(self, clusters=None, canaries=None, max_in_flight=1):  # noqa: E501
        """UpgradeClustersRequest - a model defined in Swagger"""  # noqa: E501

        self._clusters = None
        self._canaries = None
        self._max_in_flight = None
        self.discriminator = None

        if clusters is not None:
            self.clusters = clusters
        if canaries is not None:
            self.canaries = canaries
        if max_in_flight is not None:
            self.max_in_flight = max_in_flight

    @property
    def clusters(self):
        """Gets the clusters of this UpgradeClustersRequest.  # noqa: E501


        :return: The clusters of this UpgradeClustersRequest.  # noqa: E501
        :rtype: list[str]
        """
        return self._clusters

    @clusters.setter
    def clusters(self, clusters):
        """Sets the clusters of this UpgradeClustersRequest.


        :param clusters: The clusters of this UpgradeClustersRequest.  # noqa: E501
        :type: list[str]
        """

        self._clusters = clusters

    @property
    def canaries(self):
        """Gets the canaries of this UpgradeClustersRequest.  # noqa: E501


        :return: The canaries of this UpgradeClustersRequest.  # noqa: E501
        :rtype: list[str]
        """
        return self._canaries

    @canaries.setter
    def canaries(self, canaries):
        """Sets the canaries of this UpgradeClustersRequest.


        :param canaries: The canaries of this UpgradeClustersRequest.  # noqa: E501
        :type: list[str]
        """

        self._canaries = canaries

    @property
    def max_in_flight(self):
        """Gets the max_in_flight of this UpgradeClustersRequest.  # noqa: E501


        :return: The max_in_flight of this UpgradeClustersRequest.  # noqa: E501
        :rtype: int
        """
        return self._max_in_flight

    @max_in_flight.setter
    def max_in_flight(self, max_in_flight):
        """Sets the max_in_flight of this UpgradeClustersRequest.


        :param max_in_flight: The max_in_flight of this UpgradeClustersRequest.  # noqa: E501
        :type: int
        """

        self._max_in_flight = max_in_flight

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(UpgradeClustersRequest, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, UpgradeClustersRequest):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other
